#!/usr/bin/env python3

"""
    XML creator
    -----------

    This library provides custom XML tools
    for ruledxml using the lxml library.

    (C) 2015, meisterluk, BSD 3-clause license
"""

import lxml.etree

from . import exceptions


def read(xmlinfile: str):
    """Given a filepath or file descriptor to an XML file, read the XML.

    :param xmlinfile:   file descriptor to XML file
    :type xmlinfile:    str
    :return:            an object representing the document object model
    :rtype:             lxml.etree.Element
    """
    return lxml.etree.parse(xmlinfile).getroot()


def write(dom: lxml.etree.Element, fd, encoding='utf-8', **lxml_options):
    """Write a given DOM into open file descriptor `fd`.

    :param dom:             a DOM (ie. root element) to store in an XML file
    :type dom:              lxml.etree.Element
    :param fd:              the file descriptor to write to
    :type fd:               _io.TextIOWrapper
    :param encoding:        which encoding shall be used for the XML file?
    :type encoding:         str
    :param lxml_options:    options for the lxml.etree.tostring
    :type lxml_options:     dict
    """
    opts = {
        'xml_declaration': True,
        'pretty_print': True,
        'encoding': encoding
    }
    opts.update(lxml_options)
    fd.write(lxml.etree.tostring(dom, **opts))


def strip_last_element(path):
    """Strip the last element off an XPath `path`.
    Raises InvalidPathException, if path refers to an attribute.

    For example ``/root/child`` returns ``/root`` and ``child``.

    :param path:    the XPath
    :type path:     str
    """
    index = path.rfind('/')
    if index == -1:
        base, last = '', path
    else:
        base, last = path[0:index], path[index + 1:]

    if '@' in last:
        msg = 'Expected path {} to refer to an element; refers to attribute'
        raise exceptions.InvalidPathException(msg.format(path))
    else:
        return base, last


def traverse(dom, path, *,
    initial_element=lambda elem: lxml.etree.Element(elem),
    multiple_options=lambda opts: opts[0],
    no_options=lambda elem, current: None,
    finish=lambda elem, attribute='', attr_xmlns='': None) -> tuple:
    """Traverse an XPath `path` in `dom`.

    *initial_element(name)*
      Called to create the root element representing the tree
    *multiple_options(alternatives)*
      Called whenever `path` is ambiguous.
      Return value is one of `alternatives` to pick to continue traversal.
    *no_options(name, current)*
      Called whenever at element `current` in DOM,
      element `name` does not exist. Return value
      contains a new element to consider as new current
      element. If None is returned instead, traversal is aborted.
    *finish(element, attribute, attr_xmlns)*
      Called when traversal is about to finish at `element`.
      If `attribute` is non-empty, it is the attribute name requested
      in the original `path` and `attr_xmlns` attribute's namespace.
      Return value is second return value of `traverse` function.

    :param dom:                 a root node representing a DOM
    :type dom:                  lxml.etree.Element
    :param path:                an XPath to traverse
    :type path:                 str
    :param multiple_options:    see above
    :type multiple_options:     function
    :param no_options:      see above
    :type no_options:       function
    :param finish:      see above
    :type finish:       function
    :return:        A root node for the new XML DOM and the finish return value
    :rtype:         tuple([lxml.etree.Element, *])
    """
    path, *attrs = str(path).split('@')
    elements = path.lstrip('/').split('/')
    if elements[-1] == '':
        elements = elements[:-1]

    current = dom
    for i, pelement in enumerate(elements):
        if i == 0 and dom is None:
            current = dom = initial_element(pelement)
            continue
        elif i == 0 and (dom.tag == pelement or dom.tag.endswith("}" + pelement)):
            # <tag>.xpath("tag") returns []   => current = dom
            continue

        options = current.xpath(pelement)

        if len(options) == 0 or options is None:
            current = no_options(name=pelement, current=current)
            if current is None:
                return dom, None
        elif len(options) == 1:
            current = options[0]
        else:
            current = multiple_options(options)

    if attrs:
        if ':' in attrs[0]:
            xmlns, attr = attrs[0].split(':')
        else:
            xmlns, attr = None, attrs[0]

        return dom, finish(element=current, attribute=attr, xmlns=xmlns)
    else:
        return dom, finish(element=current)


def xmlns_to_lxml(element, xmlmap={}):
    """Given an `element` and an optional `xmlmap`. Return lxml-element name.

    >>> xmlns_to_lxml('name')
    'name'
    >>> xmlns_to_lxml('text', xmlmap={None: "http://www.w3.org/2000/svg",
    ... "xhtml":"http://www.w3.org/1999/xhtml"})
    ...
    '{http://www.w3.org/2000/svg}text'
    >>> xmlns_to_lxml('xhtml:text', xmlmap={None: "http://www.w3.org/2000/svg",
    ... "xhtml":"http://www.w3.org/1999/xhtml"})
    ...
    '{http://www.w3.org/1999/xhtml}text'

    :param element:     An element/element name
    :type element:      str
    :param xmlmap:      associates namespace identifiers to its URIs
    :type xmlmap:       dict
    :return:            lxml-element name (eg. useful for lxml.etree.Element)
    :rtype:             str
    """
    if not xmlmap:
        return element

    try:
        if ':' in element:
            ns, tag = element.split(":")
        else:
            ns, tag = None, element
    except ValueError:
        msg = "Invalid element name: '{}'".format(element)
        raise exceptions.InvalidPathException(msg)

    try:
        uri = xmlmap[ns]
    except KeyError:
        if ns is None:
            ns = 'default namespace'
        msg = "Unknown XML namespace: {}".format(ns)
        raise exceptions.InvalidPathException(msg)

    return '{' + uri + '}' + tag


def write_base_destination(dom: lxml.etree.Element, path: str, value,
    bases: list, xmlmap=None) -> lxml.etree.Element:
    """Behaves very much like `write_destination`, but also accepts `bases`,
    which defines a set of elements which is considered if the path is ambiguous.

    `value` will be converted to a string.

    :param dom:     root element of an XML DOM
    :type dom:      lxml.etree.Element
    :param path:    XPath to apply
    :type path:     str
    :param value:   the value to be written as text content or attribute value
    :param bases:   a set of elements considered if path is ambiguous
    :type bases:    iterable
    :param xmlmap:  Create new elements with given xmlmap and
                    traverse `path` with given `xmlmap`
    :type xmlmap:   dict
    :return:        text content, attribute or ''
    :rtype:         str
    """
    def root(name):
        return lxml.etree.Element(xmlns_to_lxml(name, xmlmap))

    def base_or_first(alternatives):
        for alt in alternatives:
            if alt in bases:
                return alt
        return alternatives[0]

    def write(element, *, attribute='', xmlns=None):
        if attribute:
            element.attrib[attribute] = str(value)
        else:
            element.text = str(value)

    def create_element(name, current):
        new_element = lxml.etree.Element(name, nsmap=xmlmap)
        current.append(new_element)
        return new_element

    return traverse(dom, path, initial_element=root,
        multiple_options=base_or_first, no_options=create_element,
        finish=write)[0]


def read_base_source(dom: lxml.etree.Element, path: str, bases: list) -> str:
    """Behaves very much like `read_source`, but also accepts `bases`, which
    defines a set of elements which is considered if the path is ambiguous.

    :param dom:     root element of an XML DOM
    :type dom:      lxml.etree.Element
    :param path:    XPath to apply
    :type path:     str
    :param bases:   a set of elements considered if path is ambiguous
    :type bases:    iterable
    :return:        text content, attribute or ''
    :rtype:         str
    """
    def base_or_first(alternatives):
        for alt in alternatives:
            if alt in bases:
                return alt
        return alternatives[0]

    def read(element, *, attribute='', xmlns=None):
        # TODO: namespace support
        if attribute:
            return str(element.attrib[attribute])
        else:
            return str(element.text or '')

    def abort(name, current):
        return None

    return traverse(dom, path, multiple_options=base_or_first,
        no_options=abort, finish=read)[1] or ''


def write_new_ambiguous_element(dom: lxml.etree.Element, path: str,
    bases=None, xmlmap=None) -> lxml.etree.Element:
    """Given a `path`, traverse it in `path`, use `bases` on ambiguous elements
    and create a new element for the top-level element of `path`.

    If a base for decision is missing, the first match is taken.

    :param dom:     root element of a DOM tree
    :type dom:      lxml.etree.Element
    :param path:    XPath to apply
    :type path:     str
    :param bases:   bases (ie. elements) to use if ambiguous
    :type bases:    list
    :param xmlmap:  Create new elements with given xmlmap and
                    traverse `path` with given `xmlmap`
    :type xmlmap:   dict
    :return:        the new created element at `path`
    :rtype:         lxml.etree.Element
    """
    path, last = strip_last_element(path)

    if bases is None:
        bases = []
    if not last:
        msg = "Path '{}' does not specify an element to create"
        raise exceptions.InvalidPathException(msg.format(path))

    def base_or_first(alternatives):
        for alt in alternatives:
            if alt in bases:
                return alt
        return alternatives[0]

    def return_element(element, *, attribute='', xmlns=None):
        if attribute:
            msg = "Expected reference to element, but attribute {} reference given"
            raise exceptions.InvalidPathException(msg.format(attribute))
        return element

    def create_element(name, current):
        new_element = lxml.etree.Element(name, nsmap=xmlmap)
        current.append(new_element)
        return new_element

    last_element = traverse(dom, path, multiple_options=base_or_first,
        no_options=create_element, finish=return_element)[1]
    new_element = create_element(last, last_element)

    return new_element


def read_ambiguous_element(dom: lxml.etree.Element, path: str, bases=None) -> list:
    """Given a `path`, traverse it in `path`, use `bases` on ambiguous elements
    and return all elements which exist at the most-nested level of `path`.

    If a base for decision is missing, the first match is taken.

    :param dom:     root element of a DOM tree
    :type dom:      lxml.etree.Element
    :param path:    XPath to apply
    :type path:     str
    :param bases:   bases (ie. elements) to use if ambiguous
    :type bases:    list
    :return:        a list of elements at `path`
    :rtype:         list([lxml.etree.Element])
    """
    path, last = strip_last_element(path)

    if bases is None:
        bases = []
    if not last:
        return dom.xpath(last)

    def base_or_first(alternatives):
        for alt in alternatives:
            if alt in bases:
                return alt
        return alternatives[0]

    def return_element(element, *, attribute='', xmlns=None):
        if attribute:
            msg = "Expected reference to element, but attribute {} reference given"
            raise exceptions.InvalidPathException(msg.format(attribute))
        return element

    def cont(name, current):
        return None

    last_element = traverse(dom, path, multiple_options=base_or_first,
        no_options=cont, finish=return_element)[1]
    if last_element is None:
        return []

    return last_element.xpath(last) or []


def write_destination(dom: lxml.etree.Element, path: str, value,
    xmlmap=None) -> lxml.etree.Element:
    """Write a `value` to an XPath `path` in `dom`.
    If `path` points to element, set text node to `value`.
    If `path` points to attribute, set attribute content to `value`.
    `value` will be converted to a string before written.

    This corresponds to the behavior of @destination.

    :param dom:     root element of an XML DOM
    :type dom:      lxml.etree.Element
    :param path:    XPath to apply
    :type path:     str
    :param value:   a value to write, string representation is taken
    :param xmlmap:  Create new elements with given xmlmap and
                    traverse `path` with given `xmlmap`
    :type xmlmap:   dict
    :return:        the (potentially modified) `dom` element
    :rtype:         lxml.etree.Element
    """
    def root(name):
        element_id = xmlns_to_lxml(name, xmlmap=xmlmap)
        return lxml.etree.Element(element_id, nsmap=xmlmap)

    def first(alternatives):
        return alternatives[0]

    def write(element, *, attribute='', xmlns=None):
        if attribute and ':' not in attribute:
            element.attrib[attribute] = str(value)
        elif attribute:
            ns, attr = attribute.split(':')
            element.attrib['{%s}%s' % (xmlns[ns], attr)] = str(value)
        else:
            element.text = str(value)

    def cont(name, current):
        new_element = lxml.etree.Element(name, nsmap=xmlmap)
        current.append(new_element)
        return new_element

    return traverse(dom, path, initial_element=root,
        multiple_options=first, no_options=cont, finish=write)[0]


def read_source(dom: lxml.etree.Element, path: str) -> str:
    """Apply a XPath `path` to `dom`. If path is ambiguous, take first option.
    If `path` points to element, return text node of it.
    If `path` points to attribute, return attribute content as string.
    If any error occurs, return an empty string.

    This corresponds to the behavior of @source.

    :param dom:     root element of an XML DOM
    :type dom:      lxml.etree.Element
    :param path:    XPath to apply
    :type path:     str
    :return:        text content, attribute or ''
    :rtype:         str
    """
    # TODO: namespace support
    if path == '':
        return ''
    elif '@' in path:
        val = dom.xpath(path)
        if val:
            return val[0] or ''
        else:
            return ''
    else:
        elements = dom.xpath(path)
        if elements:
            return elements[0].text or ''
        else:
            return ''



#!/usr/bin/env python3

"""
    ruledxml.exceptions
    -------------------

    Custom exceptions occuring at ruledxml runtime.

    ┌ RuledXmlException
    ├─── InvalidPathException [ValueError]
    ├──┬ RuleSourceException [TypeError]
    │  ├─── MissingRuleSource
    │  └─── InvalidRuleSource
    ├──┬ RuleDestinationException [TypeError]
    │  ├─── MissingRuleDestination
    │  ├─── InvalidRuleDestination
    │  └─── TooManyRuleDestinations
    └──┬ RuleForeachException [TypeError]
       ├─── MissingRuleForeach
       └─── InvalidRuleForeach

    (C) 2015, meisterluk, BSD 3-clause license
"""


class RuledXmlException(Exception):
    """The common exception base class of all custom exceptions"""


class InvalidPathException(RuledXmlException, ValueError):
    """The specified path is invalid, ie. an element has an invalid name,
    an element reference instead of an attribute reference is excepted
    or an XML namespace is unknown.
    """


class RuleSourceException(RuledXmlException, TypeError):
    """The rule has an invalid @source, missing some @source
    or too many @source calls.
    """


class RuleDestinationException(RuledXmlException, TypeError):
    """The rule has an invalid @destination, missing some @destination
    or too many @destination calls.
    """


class RuleForeachException(RuledXmlException, TypeError):
    """The rule has an invalid @foreach, missing some @foreach
    or too many @foreach calls.
    """


class MissingRuleSource(RuleSourceException):
    """The rule is lacking a @source decorator"""


class MissingRuleDestination(RuleDestinationException):
    """The rule is lacking a @destination decorator"""


class MissingRuleForeach(RuleForeachException):
    """The rule is lacking a @foreach decorator"""


class InvalidRuleSource(RuleSourceException):
    """The rule's @source decorator has invalid arguments"""


class InvalidRuleDestination(RuleDestinationException):
    """The rule's @destination decorator has invalid arguments"""


class InvalidRuleForeach(RuleForeachException):
    """The rule's @foreach decorator has invalid arguments"""


class TooManyRuleDestinations(RuleDestinationException):
    """The rule has too many @destination decorators applied"""


#!/usr/bin/env python

"""
    ruledxml
    ~~~~~~~~~
    
    Rule-based XML transformations.
    Implementation module.

    **Remarks:**
    * Function documentation is rather verbose. Use help().
    * Tested with python 3.4
    * Rules are not as complete as XSLT transformations

    (C) 2015, meisterluk, BSD 3-clause license
"""

__author__ = 'Lukas Prokop <admin@lukas-prokop.at>'
__version__ = '1.4.0'
__license__ = '3-clause BSD license'
__docformat__ = 'reStructuredText en'


# shorter names
from .core import read_rulesfile
from .xml import read as read_source_xml
from .xml import write as write_target_xml

# names with modified module ref
from .decorators import source, destination, foreach

# generic names
from .core import unique_function, required_exists
from .core import apply_rules, batch_run, run
from . import xml
from . import exceptions
from . import fs


__all__ = [
    'read_source_xml', 'read_rulesfile', 'write_target_xml',
    'source', 'destination', 'foreach',
    'unique_function', 'required_exists', 'batch_run', 'run'
    'xml', 'exceptions', 'fs'
]



#!/usr/bin/env python3

"""
    ruledxml.decorators
    -------------------

    Decorators implemented for ruledxml to be applied to rules.

    (C) 2015, meisterluk, BSD 3-clause license
"""


def annotate_function(key, init, clbk):
    """Magic function introducing a ``metadata`` dictionary attribute.
    If this attribute does not exist yet, index `key` will be set to `init`.
    Otherwise `clbk` will be executed with the old metadata attribute as argument.
    The attribute's new value is the return value of `clbk`.

    :param key:         metadata attribute name
    :type key:          str | int
    :param init:        initial value
    :param clbk:        A callback taking one argument and returning new value for key
    :type:              function
    """
    def outer(func):
        if hasattr(func, 'metadata'):
            if key in func.metadata:
                func.metadata[key] = clbk(func.metadata.get(key))
            else:
                func.metadata[key] = init
        else:
            setattr(func, 'metadata', { key : init })

        def inner(*args, **kwargs):
            return func(*args, **kwargs)

        inner.metadata = func.metadata
        return inner
    return outer


def source(*vals):
    """Decorator: Declare the source values that data will be taken from"""
    return annotate_function("src", list(vals), lambda v: v.extend(vals) or v)


def destination(*vals):
    """Decorator: Declare the destination values that data will be written to"""
    return annotate_function("dst", list(vals), lambda v: v.extend(vals) or v)


def foreach(*vals):
    """Decorator: Define a 1:1 mapping between @source and @destination.
    Let e be an element with @foreach(V, X), @source(W) and @destination(Y).
    W must be a nested element of V or V=W unless W is an attribute.
    Y must be a nested element of X or X=Y.
    @foreach only changes the behavior of @source and @destination
    iff there are multiple elements at V.

    For every element V_i, a destination element Y_i is created.
    For every element V_i,
      * the corresponding source element W is looked up and
      * provided as argument to the rule function and
      * the return value of the rule function is taken and
      * written to the corresponding destination Y
    """
    return annotate_function("each", [tuple(vals)], lambda v: v.extend([vals]) or v)



#!/usr/bin/env python

"""
    ruledxml
    ~~~~~~~~

    Rule-based XML transformations.
    To install *ruledxml* use pip3:

    .. code:: bash
        $ pip3 install ruledxml

    (C) 2015, meisterluk, BSD 3-clause license
"""

import os.path

from setuptools import setup


def readfile(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as fp:
        return fp.read()

setup(
    name='ruledxml',
    version='1.4.0',
    url='http://lukas-prokop.at/proj/ruledxml/',
    license='BSD',
    author='Lukas Prokop',
    author_email='admin@lukas-prokop.at',
    description='Rule-based XML transformations',
    long_description=readfile('README.rst'),
    packages=['ruledxml'],
    platforms='any',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Financial and Insurance Industry',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.4',
        'Topic :: Office/Business',
        'Topic :: Text Processing :: Markup :: XML'
    ],
    requires=['lxml (==3.4.4)'],
    scripts=['bin/ruledxml', 'bin/ruledxml-batched']
)



#!/usr/bin/env python3

"""
    ruledxml.fs
    -----------

    File system functionalities for ruledxml.

    (C) 2015, meisterluk, BSD 3-clause license
"""

import shutil
import string
import os.path
import pathlib
import itertools
import lxml.etree


def create_unique_filepath(folder, prefix='', suffix=''):
    """Create a filepath to a file which does not exist in `folder` yet.
    The filepath starts with `prefix` and ends with `suffix`.

    :param folder:      folder in which return value shall be unique
    :type folder:       str
    :param prefix:      prefix of the filename
    :type prefix:       str
    :param suffix:      suffix of the filename
    :type suffix:       str
    :return:            the new filepath
    :rtype:             str
    """
    def candidates():
        for i in itertools.count(1):
            for variation in itertools.product(string.ascii_lowercase, repeat=i):
                yield ''.join(variation)

    candidate = os.path.join(folder, prefix + suffix)
    for cand in candidates():
        if not os.path.exists(candidate):
            return candidate
        candidate = os.path.join(folder, prefix + '_' + cand + suffix)

    return candidate


def create_base_directories(path: str, *, wholepath=False):
    """Given a filepath, make sure all base directories exist.
    Eg. "/root/sub/file.py", make sure "/root" and "/root/sub" exists.

    :param path:        A filepath
    :type path:         str
    :param wholepath:   If True, all path elements must exist as folders;
                        in the example: also "file.py" must exist as folder
    :type wholepath:    bool
    """
    if wholepath:
        basedir = path
    else:
        basedir = os.path.split(path)[0]

    pathinst = pathlib.Path(basedir)
    if not pathinst.exists():
        pathinst.mkdir(parents=True)


def copy_files(src: list([str]), dest: str):
    """Copy files `src` to a destination folder `dest`.

    This function creates unique file paths if files with
    the same name already exist. If `dest` does not exist,
    the requested folder structure is created.

    :param src:     an iterable of source filepaths
    :type str:      list
    :param dest:    a filepath to a folder where to write files to
    :type dest:     str
    """
    create_base_directories(dest, wholepath=True)
    assert os.path.isdir(dest), "Destination path must be folder"

    for s in src:
        basename = os.path.basename(s)
        root, ext = os.path.splitext(basename)
        target = create_unique_filepath(dest, root, ext)
        shutil.copy(s, target)


def copy_file(src: str, dest: str):
    """Copy a file `src` to a destination `dest`.

    This function creates unique file paths if files with
    the same name already exist. If `dest` does not exist,
    the requested folder structure is created.

    :param src:     source filepath
    :type str:      list
    :param dest:    destination filepath
    :type dest:     str
    """
    folders, filename = os.path.split(dest)

    # create parent directories
    create_base_directories(folders, wholepath=True)

    # create unique filename
    root, ext = os.path.splitext(filename)
    target = create_unique_filepath(folders, root, ext)

    shutil.copy(src, target)



#!/usr/bin/env python3

"""
    ruledxml.core
    -------------

    Core implementation for application of rules to XML files.
    It covers the following steps:

    1. Read rules file
    2. Retrieve source XML file
    3. Do required elements exist?
    4. Apply rules
    5. Write resulting XML to file

    (C) 2015, meisterluk, BSD 3-clause license
"""

import re
import sys
import os.path
import logging
import pathlib
import argparse
import importlib.machinery
import collections

import lxml.etree

from . import fs
from . import xml
from . import exceptions


def unique_function(filepath: str):
    """Quick&dirty check that every python rule function has a unique name.

    :param filepath:            Filepath to a python file
    :type filepath:             str
    :raises RuledXmlException:  if a function name is defined twice
    """
    pattern = re.compile("def\s+(rule\w+)")
    functions = {}
    errmsg = ("Function name {} is defined multiple times "
              "in file {} (lines {} and {})")

    with open(filepath) as fp:
        for lineno, line in enumerate(fp):
            m = pattern.match(line)
            if not m:
                continue

            name = m.group(1)
            if name in functions:
                first = functions[name]
                msg = errmsg.format(name, filepath, first, lineno)
                raise exceptions.RuledXmlException(msg)

            functions[name] = lineno
            logging.info("Found {} at line {}".format(name, lineno))
            logging.debug("function definition list extended: " +
                str(functions))


def required_exists(dom: lxml.etree.Element, nonempty=None, required=None, *, filepath=''):
    """Validate `required` and `nonempty` fields.
    ie. raise InvalidPathException if path does not exist in `dom`.

    :param dom:                   the root element of a DOM to validate
    :type dom:                    lxml.etree.Element
    :param nonempty:              set of paths with nonempty values
    :type nonempty:               set
    :param required:              set of required paths
    :type required:               set
    :param filepath:              filepath (additional info for error message)
    :type filepath:               str
    :raises InvalidPathException: some required path does not exist / is empty
    """
    if not required:
        required = set()
    if not nonempty:
        nonempty = set()

    suffix = ""
    if filepath:
        suffix = " in XML file '{}'".format(filepath)

    for req in required:
        if not dom.xpath(req):
            errmsg = 'Path {} does not exist{}'.format(req, suffix)
            raise exceptions.InvalidPathException(errmsg.format(req))

    for req in nonempty:
        if xml.read_source(dom, req) == '':
            errmsg = 'Path {} is empty{}; must contain value'.format(req, suffix)
            raise exceptions.InvalidPathException(errmsg.format(req))


def read_rulesfile(filepath: str) -> tuple([dict, set]):
    """Given a `filepath`, return its contained rules and required attributes.
    Raises a exceptions.RuledXmlException if file does not contain any rule.

    :param filepath:        A filepath in the filesystem
    :type filepath:         str
    :return:                rules (associates name to implementation) and metadata
                            such as required attributes, xml namespaces and encoding
    :rtype:                 tuple(dict, dict)
    """
    def modulename(path: str) -> str:
        """Return module name for a rule file in given path"""
        return os.path.splitext(os.path.basename(path))[0]

    logging.info('Reading rules from %s', filepath)

    loader = importlib.machinery.SourceFileLoader(modulename(filepath), filepath)
    rulesfile = loader.load_module()

    rules = {}
    required = set()
    nonempty = set()
    encoding = "utf-8"
    xml_namespaces = {}

    for member in dir(rulesfile):
        if member.startswith("rule"):
            rules[member] = getattr(rulesfile, member)
            logging.info("Found %s", member)
        elif member == "required":
            required = set(getattr(rulesfile, member))
            logging.info("Found required attribute with %d elements", len(required))
        elif member == "nonempty":
            nonempty = set(getattr(rulesfile, member))
            logging.info("Found nonempty attribute with %d elements", len(nonempty))
        elif member.endswith("namespaces"):
            xml_namespaces = getattr(rulesfile, member)
        elif member == "encoding":
            encoding = getattr(rulesfile, member)

    if not rules:
        msg = "Expected at least one rule definition, none given in {}"
        raise exceptions.RuledXmlException(msg.format(filepath))

    msg = '{!r} {!r} {!r} {!r}'
    logging.debug(msg.format(rules, required, xml_namespaces, encoding))
    logging.info('%d rules, %d required elements and %d XML namespaces',
        len(rules), len(required), len(xml_namespaces))
    logging.info('encoding is set to {}'.format(encoding))

    metadata = {
        'required': required,
        'nonempty': nonempty,
        'xmlmap': xml_namespaces,
        'encoding': encoding
    }
    return rules, metadata


def validate_rules(rules: dict):
    """Validate rules. Test whether decorator setup is fine for all of them.

    :param rules:               rule names associated to their implementation
    :type rules:                dict(str: function)
    :raises RuledXmlException:  a decorator is set up wrongfully
    """
    for rulename, rule in rules.items():
        # a rule must have @source, @destination or @foreach applied
        if not hasattr(rule, 'metadata'):
            msg = ("Function {} is considered to be a rule, but this "
                   "requires at least a @destination declaration")
            raise exceptions.InvalidRuleDestination(msg.format(rulename))

        # a rule must have at least one @destination
        dst_len = len(rule.metadata.get('dst', []))
        if dst_len != 1:
            msg = "A rule must have exactly 1 @destination. {} has {}"
            raise exceptions.TooManyRuleDestinations(msg.format(rulename, dst_len))

        # distinguish: foreach, no-foreach
        if 'each' in rule.metadata:
            each_len = len(rule.metadata['each'])
            if each_len == 0:
                msg = "A @foreach rule requires at least 2 arguments. {} has 0"
                raise exceptions.InvalidRuleForeach(msg.format(rulename))

            each_arg_len = len(rule.metadata['each'][0])
            if each_arg_len != 2:
                msg = "@foreach must have exactly two arguments. {} has {}"
                raise exceptions.InvalidRuleForeach(msg.format(rulename, each_arg_len))

            each_dst_len = len(rule.metadata.get('dst', []))
            if each_dst_len > 1:
                msg = ("@foreach rules must have at most "
                       "1 @destination. {} has {}")
                raise exceptions.InvalidRuleForeach(msg.format(rulename, each_dst_len))

            # outer @foreach[0] must be prefix of innner @foreach[0]
            destination = rule.metadata['dst'][0]
            for source in rule.metadata.get('src', []):
                prev_base_src = None
                for base_src, base_dst in rule.metadata['each']:
                    if prev_base_src is not None and not base_src.startswith(prev_base_src):
                        msg = ("Outer first @foreach argument '{}' must be prefix of "
                               "inner first @foreach argument '{}'")
                        raise exceptions.InvalidRuleForeach(msg.format(prev_base_src, base_src))

                    prev_base_src = base_src

        else:
            pass  # no further checks


def build_recursive_structure(rules: list) -> list:
    """Build a recursive structure based on foreach-sources of the given rules.

    >>> rule1 = {
    ...   'foreach': [('a', 'x'), ('ab', 'xy')],
    ...   'source': ['1', 'a2', 'ab3', 'ab4'],
    ...   'destination': ['xy5']
    ... }
    >>> rule2 = {
    ...   'foreach': [('a', 'x'), ('ac', 'xz')],
    ...   'source': ['6'],
    ...   'destination': ['xz7']
    ... }
    >>> build_recursive_structure([rule1, rule2])
    [{
        'class': 'iteration',
        'dstbase': 'x',
        'srcbase': 'a',
        'children': [
            {   'class': 'iteration',
                'dstbase': 'xy',
                'srcbase': 'ab',
                'children': []
            }, {'class': 'iteration',
                'dstbase': 'xz',
                'srcbase': 'ac',
                'children': []
        }]
    }]

    :param rules:       a set of rules to read @foreach attributes from
    :type rules:        list
    :return:            a list of (potentially nested) dictionaries
    :rtype:             list
    """
    all_each_bases = set()
    for rule in rules:
        for each in rule['foreach']:
            all_each_bases.add(each)

    all_each_bases = list(all_each_bases)
    all_each_bases.sort(key=lambda e: e[0])

    structure = []

    for base in all_each_bases:
        current = structure

        added = False
        while not added:
            for element in current:
                if base[0].startswith(element['srcbase']):
                    current = element['children']
                    break
            else:
                childs = []
                current.append({
                    'class': 'iteration',
                    'srcbase': base[0],
                    'dstbase': base[1],
                    'children': childs
                })
                current = childs
                added = True

    return structure


def classify_rules(rules: dict):
    """Classify rules. Represent rules as dictionary with associated metadata.
    Returns a data structure with is nicely structured to perform the @source
    and @destination algorithms with respect to @foreach semantics.

    :param rules:       rule names associated to their implementation
    :type rules:        dict(str: function)
    :return:            a list of dictionaries containing rules with metadata;
                        might be recursive (dicts contain lists of dicts)
    :rtype:             [dict(), dict(), ...]
    """
    # I store an 'order' attribute.
    # It is not used for processed, only for documentation.
    # Lists are ordered anyway and elements are added in the correct order.

    classified = []
    orders = set(range(len(rules)))

    # add basic rules
    each_found = False
    for rulename, rule in rules.items():
        if 'each' in rule.metadata:
            each_found = True
            continue

        order = min(orders)
        orders.remove(order)

        classified.append({
            'name': rulename,
            'class': 'basicrule',
            'rule': rule,
            'each': [],
            'src': rule.metadata.get('src', []),
            'dst': rule.metadata.get('dst', []),
            'order': (order,)
        })

    if not each_found:
        return classified

    # collect the set of all @foreach base sources
    foreach_rules = []
    for rulename, rule in rules.items():
        if 'each' not in rule.metadata:
            continue

        foreach_rules.append({
            'foreach': rule.metadata.get('each', []),
            'source': rule.metadata.get('src', []),
            'destination': rule.metadata.get('dst', [])
        })

    # build recursive structure for @foreach entries
    # node tell when an ambiguous element has to be iterated
    recursive_structure = build_recursive_structure(foreach_rules)

    # annotate rules to it
    def traverse(tree, xpath):
        # Assumption. xpath exists as base in tree.
        assert xpath
        current = tree
        found = False
        while not found:
            for element in current:
                if xpath == element['srcbase']:
                    return element['children']
                if xpath.startswith(element['srcbase']):
                    current = element['children']

    # add the rules to the recursive structure
    for rulename, rule in rules.items():
        if 'each' not in rule.metadata:
            continue
        
        most_nested = rule.metadata['each'][-1]
        lst = traverse(recursive_structure, most_nested[0])
        lst.append({
            'class': 'foreach-rule',
            'name': rulename,
            'src': rule.metadata.get('src', []),
            'dst': rule.metadata.get('dst', []),
            'rule': rule
        })

    for struct in recursive_structure:
        order = min(orders)
        orders.remove(order)
        struct['order'] = order

        classified.append(struct)

    return classified


def run_rules(src_dom: lxml.etree.Element, target_dom: lxml.etree.Element,
    classified: list, xmlmap=None):
    """Actually apply the classified rules to a target DOM.

    :param src_dom:     the root element of a DOM to retrieve source data from
    :type src_dom:      lxml.etree.Element
    :param target_dom:  the root element of a DOM to write destination data to
    :type target_dom:   lxml.etree.Element
    :param classified:  a list of dictionaries containing rules with metadata;
                        might be recursive (dicts contain lists of dicts)
    :type classified:   [dict(), dict(), ...]
    :param xmlmap:      association of XML namespace name to URI
    :type xmlmap:       dict
    :param bases:       Base elements (created for source @foreach elements)
    :type bases:        list
    :return:            the root element of a new DOM
    :rtype:             lxml.etree.Element
    """
    def finish_a_tree(src_dom, target_dom, node, src_bases, dst_bases):
        if node['class'] == 'iteration':
            for src_base in xml.read_ambiguous_element(src_dom, node['srcbase'], src_bases):
                dst_base = xml.write_new_ambiguous_element(target_dom,
                    node['dstbase'], dst_bases, xmlmap)
                for child in node['children']:
                    target_dom = finish_a_tree(src_dom, target_dom, child,
                        src_bases.copy() + [src_base], dst_bases.copy() + [dst_base])
            return target_dom
        elif node['class'] == 'foreach-rule':
            args = []
            for src in node['src']:
                args.append(xml.read_base_source(src_dom, src, bases=src_bases))
            output = node['rule'](*args)
            if output is None:
                return target_dom
            return xml.write_base_destination(target_dom, node['dst'][0],
                output, bases=dst_bases, xmlmap=xmlmap)

    for obj in classified:
        if obj['class'] == 'basicrule':
            logging.info("Applying %s", obj['name'])

            args = []
            for src in obj['src']:
                args.append(xml.read_source(src_dom, src))

            logging.debug("Applying %s with arguments %s", obj['name'], str(args))

            output = obj['rule'](*args)
            if output is None:
                continue
            dst = obj['dst'][0]
            target_dom = xml.write_destination(target_dom, dst, output, xmlmap=xmlmap)

        elif obj['class'] in ('iteration', 'foreach-rule'):
            target_dom = finish_a_tree(src_dom, target_dom, obj, [], [])

    return target_dom


def apply_rules(dom: lxml.etree.Element, rules: dict, *, xmlmap=None):
    """Apply given rules to the given DOM.

    :param dom:                 the root element of a DOM
    :type dom:                  lxml.etree.Element
    :param rules:               rule names associated to their implementation
    :type rules:                dict(str : function)
    :param xmlmap:              association of XML namespace name to URI
    :type xmlmap:               dict
    :return:                    root element of a new DOM
    :rtype:                     lxml.etree.Element
    :raises RuledXmlException:  some rule is invalid
    """
    validate_rules(rules)
    classified = classify_rules(rules)
    return run_rules(dom, None, classified, xmlmap)


def run(in_fd, rules_filepath: str, out_df, *, infile='', outfile='') -> int:
    """Process one file.

    :param in_fd:           File descriptor to one input XML file
    :type in_fd:            _io.TextIOWrapper
    :param rules_filepath:  Filepath to a rulesfile
    :type rules_filepath:   str
    :param out_fd:          File descriptor to an output XML file
    :type out_fd:           _io.TextIOWrapper
    :param infile:          original XML input file path for debugging purposes
    :type infile:           str
    :param outfile:         output XML file path for debugging purposes
    :type outfile:          str
    :return:                exit code 0
    :rtype:                 int
    """
    # read rules file
    unique_function(rules_filepath)
    rules, meta = read_rulesfile(rules_filepath)

    # retrieve source xmlfile
    src_dom = xml.read(in_fd)

    # test: required elements exist?
    required_exists(src_dom, meta['nonempty'], meta['required'], filepath=infile)

    # apply rules
    target_dom = apply_rules(src_dom, rules, xmlmap=meta['xmlmap'])

    # write target XML to file
    xml.write(target_dom, out_df, encoding=meta['encoding'])

    return 0


def batch_run(in_fd, rules_filepath: str, out_filepaths: list([str]),
    base: str, *, infile='') -> int:
    """Process one file. Apply rules for some base path.
    Create several target DOMs.

    :param in_fd:           File descriptor to one input XML file
    :type in_fd:            _io.TextIOWrapper
    :param rules_filepath:  Filepath to a rulesfile
    :type rules_filepath:   str
    :param out_filepaths:   File descriptor to an output XML file
    :type out_filepaths:    _io.TextIOWrapper
    :param base:            A base XPath, all rules are applied relative to this path
    :type base:             str
    :param infile:          original XML input file path for debugging purposes
    :type infile:           str
    :return:                exit code 0
    :rtype:                 int
    """
    # read rules file
    unique_function(rules_filepath)
    rules, meta = read_rulesfile(rules_filepath)

    # retrieve source xmlfile
    src_dom = xml.read(in_fd)

    count = 0
    for element in src_dom.xpath(base):
        # test: required elements exist?
        required_exists(element, meta['nonempty'], meta['required'], filepath=infile)

        # apply rules
        target_dom = apply_rules(element, rules, xmlmap=meta['xmlns'])

        # write target XML to file
        fs.create_base_directories(out_filepaths[count])
        with open(out_filepaths[count], 'wb') as out_fd:
            xml.write(target_dom, out_fd, encoding=meta['encoding'])

        count += 1

    if count < len(out_filepaths):
        msg = "Number of output filepaths was {}; expected {}"
        logging.warn(msg.format(count, len(out_filepaths)))

    return 0
